/*-
 * Copyright (C)2003..2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2003..2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: open.c,v 1.31 2014/08/30 12:40:24 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT 3

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <babolo/BLINflag.h>
#include "mife_int.h"

#define TRUFLAG(A) ( (A)                                                                              \
                   & ( MIFE_STRI                                                                      \
                     | MIFE_RTRY                                                                      \
                     | MIFE_BUFL                                                                      \
                     | MIFE_SLID                                                                      \
                     | MIFE_EOFL                                                                      \
                     | MIFE_FULL                                                                      \
                     | MIFE_BUFX                                                                      \
                     | MIFE_BACK                                                                      \
                     | BLIN_MASK                                                                      \
                   ) )

BLIN_flag mife_default = BLIN_BIT0;
ssize_t   mife_read __P((mife_descriptor *mife, ssize_t ordlen, off_t offset));

static ssize_t
mindelta(BLIN_flag flags) {
    ssize_t s = 0;

    if  (flags & MIFE_BUFL) s = 1 << (((flags & MIFE_BUFL) >> MIFE_BUFS) + 16);
    if  (s < MINDELTA) s = MINDELTA;
    return(s);
}

mife_descriptor *
mife_init(BLIN_flag flags) {
    mife_descriptor *mife = NULL;

#   define blin_internal_flags (flags & BLIN_MASK)
    ifBLIN_QX3("+ flags=%08X", flags);
    SETBDEF(flags);
    if  (!(mife = calloc(1, sizeof(mife_descriptor)))) {
        ifBLIN_QW0("no mem #1");
        errno = ENOMEM;
        goto out;
    }
    mife->flags = TRUFLAG(flags);
    mife->mindelta = mindelta(mife->flags);
    if  (!(MIFE_RTRY & flags)) mife->flags |= NRETRY00;
out:
    ifBLIN_QX3("- %"BLIN_X" flags=%08X", BLIN_I(mife), mife->flags);
    return(mife);
#   undef blin_internal_flags
}

static int
setime(mife_descriptor *mife, struct timespec *stime, time_t sec, long nsec) {
    int ex = EX_OK;

    if  (!mife->kev && !(mife->kev = calloc(1, sizeof(mife_event)))) {
        ifBLIN_QX0("malloc");
        ERROUT(-EX_IOERR, ENOMEM);
    }
    if  (0 > (mife->kev->kq = kqueue())) {
        ifBLIN_QW0("kqueue");
        EXOUT(-EX_IOERR);
    }
    stime->tv_sec = sec;
    stime->tv_nsec = nsec;
out:
    return(ex);
}

ssize_t
mife_ctl(mife_descriptor *mife, int type, ...) {
    ssize_t ordlen, ex = EX_OK;
    off_t offset;
    va_list ap;
    char *s;
    int i;

#   define blin_internal_flags ((mife ? mife->flags : mife_default) & BLIN_MASK)
    ifBLIN_QX3("+ %"BLIN_X" type=%d", BLIN_I(mife), type);
    va_start(ap, type);
    if  (!mife) {
        ifBLIN_QX0("mife NULL");
        ERROUT(-EX_DATAERR, EFAULT);
    }
    switch (type) {
    case MIFE_CTLFLAG: {
            BLIN_flag flags = va_arg(ap, BLIN_flag);

            ifBLIN_QX4("  CTLFLAG %08X", flags);
            mife->flags = TRUFLAG(flags);
            if  (!(MIFE_RTRY & flags)) mife->flags |= NRETRY00;
        }
        break;
    case MIFE_CTLFILE: {
            s = va_arg(ap, char *);
            ifBLIN_QX4("  CTLFILE %s", s);
            mife->flags = TRUFLAG(mife->flags) | MIFE_CLOS;
            if  (0 > (mife->f.fd = open(s, O_RDONLY))) {
                i = errno;
                ifBLIN_QW0("open %s", s);
                mife_ctl(mife, MIFE_CTLFINI);
                mife = NULL;
                errno = i;
                EXOUT(-EX_IOERR);
            }
            mife->f.read = read;
            mife->f.close = close;
            mife->actbeg = 0;
            mife->actlen = 0;
            mife->offset = 0;
            mife->ofpoint = 0;
            mife->st_size = 0;
            if  (mife->eol) mife->eol->stop(mife);
        }
        break;
    case MIFE_CTLFDSC: {
            i = va_arg(ap, int);
            ifBLIN_QX4("  CTLFDSC %d", i);
            mife->flags = TRUFLAG(mife->flags);
            mife->f.fd = i;
            mife->f.read = read;
            mife->f.close = close;
            mife->actbeg = 0;
            mife->actlen = 0;
            mife->offset = 0;
            mife->ofpoint = 0;
            mife->st_size = 0;
            if  (mife->eol) mife->eol->stop(mife);
        }
        break;
    case MIFE_CTLCNST: {
            s = va_arg(ap, void *);
            mife->actlen = va_arg(ap, size_t);
            ifBLIN_QX4("  CTLCNST %"BLIN_X"["BLIN_U"]", BLIN_I(s), mife->actlen);
            mife->flags = TRUFLAG(mife->flags) | MIFE_INMM | MIFE_ALLC | MIFE_EOFL;
            mife->mindelta = 0;
            mife->a.ad = (void *)-1;
            mife->ofpoint = mife->offset = 0;
            mife->actbeg = 0;
            i = (mife->flags & MIFE_FULL) ? 1 : 0;
            if  (!mife->actlen && !(mife->flags & MIFE_STRI)) mife->actlen = strlen(s);
            mife->buflen = mife->actlen + i;
            if  (!(mife->buffer = malloc(mife->buflen))) {
                ifBLIN_QX0("malloc MIFE_CTLCNST");
                mife_ctl(mife, MIFE_CTLFINI);
                mife = NULL;
                ERROUT(-EX_OSERR, ENOMEM);
            }
            bcopy(s, mife->buffer, mife->actlen);
            if  (i) CHARBUF[mife->actlen] = 0;
        }
        break;
    case MIFE_CTLBUFF: {
            s = va_arg(ap, void *);
            mife->actlen = va_arg(ap, size_t);
            ifBLIN_QX4("  CTLBUFF %"BLIN_X"["BLIN_U"]", BLIN_I(s), mife->actlen);
            mife->flags = TRUFLAG(mife->flags) | MIFE_INMM | MIFE_EOFL;
            mife->mindelta = 0;
            mife->a.ad = (void *)-1;
            mife->ofpoint = mife->offset = 0;
            mife->actbeg = 0;
            i = (mife->flags & MIFE_FULL) ? 1 : 0;
            if  (!mife->actlen && !(mife->flags & MIFE_STRI)) {
                mife->actlen = strlen(s);
                mife->buflen = mife->actlen + i;
                i = 0;
            } else {
                mife->buflen = mife->actlen + i;
            }
            mife->buffer = s;
            if  (i) CHARBUF[mife->actlen] = 0;
        }
        break;
    case MIFE_CTLEOLI: {
            s = va_arg(ap, char *);
            if  (!s) {
                ifBLIN_QX4("  CTLEOLI NULL");
                if  (mife->eol) {
                    if  (mife->eol->trilen && mife->eol->trix) {
                        free(mife->eol->trix);
                        mife->eol->trix = NULL;
                    }
                    free(mife->eol);
                    mife->eol = NULL;
                }
                goto out;
            }
            ifBLIN_QX4("  CTLEOLI =%s~", s);
            if  (!(mife->eol = malloc(sizeof(mife_eol)))) {
                ifBLIN_QX0("malloc MIFE_CTLEOLI");
                ERROUT(-EX_OSERR, ENOMEM);
            }
            bcopy(&mife_eol0, mife->eol, sizeof(mife_eol));
            if  (mife->eol->init && ((ex = mife->eol->init(mife, s)) < 0)) {
                ifBLIN_QW0("eol->init MIFE_CTLEOLI");
                goto out;
        }   }
        break;
#if defined(MIFE_CTLITIM)
    case MIFE_CTLITIM: {
#else
    case 6: {
#endif
            const struct timespec *timeout = va_arg(ap, struct timespec *);

            ifBLIN_QX4("  CTLITIM %d.%09d", timeout->tv_sec, timeout->tv_nsec);
            ex = setime(mife, &mife->itime, timeout->tv_sec, timeout->tv_nsec);
        }
        break;
#if defined(MIFE_CTLOTIM)
    case MIFE_CTLOTIM: {
#else
    case 7: {
#endif
            const struct timespec *timeout = va_arg(ap, struct timespec *);

            ifBLIN_QX4("  CTLOTIM %d.%09d", timeout->tv_sec, timeout->tv_nsec);
            ex = setime(mife, &mife->otime, timeout->tv_sec, timeout->tv_nsec);
        }
        break;
    case MIFE_CTLTIM0: {
            mife_tim timeout = va_arg(ap, mife_tim);

            ifBLIN_QX4("  CTLTIM0 %"BLIN_O"u", timeout);
            ex = setime(mife, &mife->itime, timeout / 1000000000ULL, timeout % 1000000000ULL);
        }
        break;
    case MIFE_CTLTIM1: {
            mife_tim timeout = va_arg(ap, mife_tim);

            ifBLIN_QX4("  CTLTIM1 %"BLIN_O"u", timeout);
            ex = setime(mife, &mife->otime, timeout / 1000000000ULL, timeout % 1000000000ULL);
        }
        break;
    case MIFE_CTLFINI: {
            ifBLIN_QX4("  CTLFINI");
            if  (mife->kev) {
                free(mife->kev);
                mife->kev = NULL;
            }
            if  (mife->eol) {
                if  (mife->eol->trilen && mife->eol->trix) {
                    free(mife->eol->trix);
                    mife->eol->trix = NULL;
                }
                free(mife->eol);
                mife->eol = NULL;
            }
            if  (mife->buffer) {
                if  ((mife->flags & MIFE_MMAP) && (ex = munmap(mife->buffer, mife->buflen))) {
                    ifBLIN_QW0("munmap");
                }
                if  (mife->flags & MIFE_ALLC) free(mife->buffer);
                mife->buffer = NULL;
                mife->buflen = 0;
            }
            if  ((mife->flags & MIFE_CLOS) && (ex = mife->a.close(mife->a.ad))) {
                ifBLIN_QW0("mife->a.close");
            }
            free(mife);
            mife = NULL;
        }
        break;
    case MIFE_CTLTEST: {
            ifBLIN_QX4( "  CTLTEST %"BLIN_X"->%"BLIN_X"[%"BLIN_D"] actbeg=%"BLIN_D" actlen=%"BLIN_D
                        " offset=%"BLIN_O"d ofpoint=%"BLIN_O"d"
                      , BLIN_I(mife)
                      , BLIN_I(mife->buffer)
                      , mife->buflen
                      , mife->actbeg
                      , mife->actlen
                      , mife->offset
                      , mife->ofpoint
                      );
            ex = -EX_DATAERR;
            if  (!mife->buffer) {
                ifBLIN_QX0("mife->buffer NULL");
                errno = EFAULT;
                goto out;
            }
            if  (!mife->buflen) {
                ifBLIN_QX0("mife->buflen 0");
                errno = EBADF;
                goto out;
            }
            if  ((mife->actbeg + mife->actlen) > mife->buflen) {
                ifBLIN_QX0("mife->actbeg %"BLIN_U" + mife->actlen %"BLIN_U" > mife->buflen %"BLIN_U
                          , mife->actbeg, mife->actlen, mife->buflen
                          );
                errno = EBADF;
                goto out;
            }
            if  (mife->flags & MIFE_FULL) {
                if  ((mife->actbeg + mife->actlen) >= mife->buflen) {
                    ifBLIN_QX0("no space for 0 in FULL");
                    errno = EBADF;
                    goto out;
                }
                if  (CHARBUF[mife->actbeg + mife->actlen]) {
                    ifBLIN_QX0("no 0 in FULL");
                    ex = EX_DATAERR;
                    errno = EINVAL;
                    goto out;
            }   }
            if  (0 > mife->offset) {
                ifBLIN_QX0("0 > mife->offset %"BLIN_O"d", mife->offset);
                errno = EBADF;
                goto out;
            }
            if  (0 > mife->ofpoint) {
                ifBLIN_QX0("0 > mife->ofpoint %"BLIN_O"d", mife->ofpoint);
                errno = EBADF;
                goto out;
            }
            if  (mife->offset > mife->ofpoint) {
                ifBLIN_QX0( "mife->offset %"BLIN_O"d > mife->ofpoint %"BLIN_O"d"
                          , mife->offset, mife->ofpoint
                          );
                errno = EBADF;
                goto out;
            }
            if  ((mife->ofpoint - mife->offset) > (off_t)mife->actlen) {
                ifBLIN_QX0( "mife->thepoint %"BLIN_O"d > mife->actlen %"BLIN_U
                          , mife->ofpoint - mife->offset, mife->actlen
                          );
                ex = EX_SOFTWARE;
                errno = EINVAL;
                goto out;
            }
            if  (mife->flags & MIFE_MMAP) {
                if  (0 > mife->st_size) {
                    ifBLIN_QX0("0 > mife->st_size %"BLIN_O"d", mife->st_size);
                    errno = EINVAL;
                    goto out;
                }
                if  (mife->ofpoint > mife->st_size) {
                    ifBLIN_QX0( "mife->ofpoint %"BLIN_O"d > mife->st_size %"BLIN_O"d"
                              , mife->ofpoint, mife->st_size
                              );
                    errno = EINVAL;
                    goto out;
            }   }
            ex = EX_OK;
            if  (!(mife->flags & (MIFE_MMAP | MIFE_PIPE | MIFE_INMM))) {
                ifBLIN_QX1("Noninitialized");
        }   }
        break;
    case MIFE_CTLPIPE: {
            int fd[2];

            ifBLIN_QX4("  CTLPIPE");
            if  (0 > (ex = mife_ctl(mife, MIFE_CTLTEST))) {
                ifBLIN_QW0("Illegal mife");
                goto out;
            } else if (!!ex) {
                ifBLIN_QX0("Incorrect mife");
                errno = EINVAL;
                ex = -ex;
                goto out;
            }
            if  (0 > pipe(fd)) {
                ifBLIN_QW0("pipe");
                ex = -EX_IOERR;
                goto out;
            }
            if  (0 > (i = fork())) {
                ifBLIN_QW0("fork");
                ex = -EX_OSERR;
                goto out;
            }
            if  (!i) {
                off_t delta, of;
                ssize_t l;

                for (l = 1, of = mife->ofpoint; l > 0;) {
                    if  (0 > (l = mife_read(mife, 0, of))) {
                        ifBLIN_QW0("mife_read");
                        ex = -EX_IOERR;
                        break;
                    }
                    of += l;
                    delta = mife->ofpoint - mife->offset;
                    l = mife_writ(fd[1], mife->buffer + mife->actbeg + delta, mife->actlen - delta);
                    if  (0 > l) {
                        ifBLIN_QW0("mife_writ");
                        ex = -EX_IOERR;
                        break;
                }   }
                exit(ex);
            }
            ex = fd[0];
        }
        break;
    case MIFE_CTLSIZE: {
            ifBLIN_QX4("  CTLSIZE");
            if  (0 > (ex = mife_ctl(mife, MIFE_CTLTEST))) {
                ifBLIN_QW0("Illegal mife");
                goto out;
            } else if (!!ex) {
                ifBLIN_QX0("Incorrect mife");
                errno = EINVAL;
                ex = -ex;
                goto out;
            }
            ex = mife->offset + mife->actlen - mife->ofpoint;
        }
        break;
    case MIFE_CTLREAD: {
            offset = va_arg(ap, off_t);
            ordlen = va_arg(ap, ssize_t);
            ifBLIN_QX4("  CTLREAD %"BLIN_O"d[%"BLIN_U"]", offset, ordlen);
            if  (0 > (ex = mife_ctl(mife, MIFE_CTLOFFT, offset))) {
                ifBLIN_QW0("mife_ctl MIFE_CTLOFFT");
                goto out;
            }
            ex = mife_read(mife, ordlen, offset);
        }
        break;
    case MIFE_CTLOFFT: {
            offset = va_arg(ap, off_t);
            ifBLIN_QX4("  CTLOFFT %"BLIN_O"d", offset);
            ex = -EX_IOERR;
            if  (0 > offset) {
                ifBLIN_QX0("0 > offset %"BLIN_O"d", offset);
                errno = EINVAL;
                goto out;
            }
            if  (mife->offset > offset) {
                ifBLIN_QX0("mife->offset %"BLIN_O"d > offset %"BLIN_O"d", mife->offset, offset);
                errno = ESPIPE;
                goto out;
            }
            if  ((mife->flags & MIFE_MMAP) && (offset > mife->st_size)) {
                ifBLIN_QX0("offset %"BLIN_O"d > st_size %"BLIN_O"d", offset, mife->st_size);
                errno = EINVAL;
                goto out;
            }
            ex = EX_IOERR;
            if  ((offset - mife->offset) > (off_t)mife->actlen) {
                ifBLIN_QX0( "offset - mife->offset %"BLIN_O"d-%"BLIN_O"d=%"BLIN_O"d "
                            "> mife->actlen %"BLIN_U
                          , offset, mife->offset, offset - mife->offset, mife->actlen
                          );
                errno = EINVAL;
                goto out;
            }
            ex = EX_OK;
        }
        break;
    default:
        ifBLIN_QX0("Illegal command %d", type);
        ex = -EX_DATAERR;
        errno = EINVAL;
    }
out:
    va_end(ap);
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

mife_descriptor *
mife_open(BLIN_flag flags, const char *path) {
    mife_descriptor *mife = NULL;
    int er;

    if  ((mife = mife_init(flags)) && mife_ctl(mife, MIFE_CTLFILE, path)) {
        er = errno;
        mife_ctl(mife, MIFE_CTLFINI);
        mife = NULL;
        errno = er;
    }
    return(mife);
}

mife_descriptor *
mife_opef(BLIN_flag flags, int fd) {
    mife_descriptor *mife = NULL;
    int er;

    if  ((mife = mife_init(flags)) && mife_ctl(mife, MIFE_CTLFDSC, fd)) {
        er = errno;
        mife_ctl(mife, MIFE_CTLFINI);
        mife = NULL;
        errno = er;
    }
    return(mife);
}

mife_descriptor *
mife_opes(BLIN_flag flags, const char *src, size_t size) {
    mife_descriptor *mife = NULL;
    int er;

    if  ((mife = mife_init(flags)) && mife_ctl(mife, MIFE_CTLCNST, src, size)) {
        er = errno;
        mife_ctl(mife, MIFE_CTLFINI);
        mife = NULL;
        errno = er;
    }
    return(mife);
}

mife_descriptor *
mife_opem(BLIN_flag flags, char *src, size_t size) {
    mife_descriptor *mife = NULL;
    int er;

    if  ((mife = mife_init(flags)) && mife_ctl(mife, MIFE_CTLBUFF, src, size)) {
        er = errno;
        mife_ctl(mife, MIFE_CTLFINI);
        mife = NULL;
        errno = er;
    }
    return(mife);
}

int
mife_fini(mife_descriptor *mife) {
    return(mife_ctl(mife, MIFE_CTLFINI));
}

__weak_reference(mife_fini, mife_close);
